"""Compatibility wrapper that translates Internermarke SOAP API calls to REST ones.

Use case: You have existing software that interfaces with the Deutsche Post
Internetmarke SOAP API via the :class:`inema.inema.Internetmarke` class, you
need to upgrade to the new Deutsche Post Internetmarke REST API (because the
SOAP endpoints are announced to be turned off at the end of 2025) and you want
to change your existing software as little as possible.

In the best case you just need to register for the new REST API, link your
Portokasse in Deutsche Post Web UI, update your credentials configuration and
adjust your import like this::

    from inema import Internetmarke

to something like this::

    from inema.soapfree import Internetmarke

See also `frank.py` in the python-inema repository (around 2025-09-21) for a
real world example.


Of course, depending on your existing usage, directly migrating to the new REST
API via :mod:`inema.rest` likely isn't that complicated, either.

"""

# SPDX-License-Identifier: LGPL-3.0-or-later
# SPDX-FileCopyrightText: © 2025 Georg Sauthoff <mail@gms.tf>


from .rest import Session, mk_addr, mk_pdf_pos, mk_png_pos, mk_pdf_req, mk_pdf_preview_req, mk_png_preview_req
from .data import products as default_products

import collections
import decimal
import io
import logging
import zipfile


log = logging.getLogger(__name__)


def translate_layout(layout):
    match layout:
        case 'AddressZone':
            layout = 'ADDRESS_ZONE'
        case 'FrankingZone':
            layout = 'FRANKING_ZONE'
    return layout


class Internetmarke:

    def __init__(self, partner_id, key, key_phase="1", products=None):
        self.partner_id = partner_id
        self.key = key
        self.positions = []
        if products is None:
            self.products = default_products
        else:
            self.products = products

    def authenticate(self, username, password):
        self.im = Session(self.partner_id, self.key, username, password)
        self.user_token =  self.im.auth_resp['access_token']
        self.wallet_balance = self.im.balance
        log.debug(f'balance: {self.wallet_balance}')

    def build_addr(self, street, house, zipcode, city, country, additional = None):
        line = f'{street} {house}'
        return (line, zipcode, city, country, additional)


    def build_pers_addr(self, first, last, address, salutation = None, title = None):
        ns = []
        if salutation is not None:
            ns.append(salutation)
        if title is not None:
            ns.append(title)
        ns.append(first)
        ns.append(last)
        name = ' '.join(ns)
        line, zipcode, city, country, additional = address
        a = mk_addr(name, line, zipcode, city, country, line2=additional)
        return a

    def build_comp_addr(self, company, address, person = None):
        if person is None:
            header = None
            name = company
        else:
            header = company
            name = person
        line, zipcode, city, country, additional = address
        a = mk_addr(name, line, zipcode, city, country, header=header, line2=additional)
        return a


    def build_position(self, product, sender=None, receiver=None, layout="AddressZone", pdf=False, x=1, y=1, page=1):
        layout = translate_layout(layout)
        if pdf:
            return mk_pdf_pos(product, x, y, page, layout=layout, sender=sender, receiver=receiver)
        else:
            return mk_png_pos(product, layout=layout, sender=sender, receiver=receiver)

    def add_position(self, position):
        self.positions.append(position)

    def clear_positions(self):
        self.positions = []

    def compute_total(self):
        total = 0
        for p in self.positions:
            total += self.get_product_price_by_id(p['productCode'])
        return total

    def get_product_price_by_id(self, ext_prod_id):
        price_str = self.products[str(ext_prod_id)]['cost_price']
        return int(decimal.Decimal(price_str) * 100)


    def checkoutPDF(self, page_format):
        oid = self.im.create_order()
        log.debug(f'new order id: {oid}')
        t = self.compute_total()
        b = mk_pdf_req(oid, page_format, self.positions, t, manifest=True)
        d = self.im.checkout_pdf(b)
        h = {}
        if d.get('link'):
            h['pdf_bin'] = self.im.ses.get(d['link']).content
        if d.get('manifestLink'):
            h['manifest_pdf_bin'] = self.im.ses.get(d['manifestLink']).content
        Tuple = collections.namedtuple('Checkout', h)
        r = Tuple(**h)
        return r

    def retrievePNGs(self, link):
        r = self.im.ses.get(link)
        z = zipfile.ZipFile(io.BytesIO(r.content))
        return [ z.read(f.filename) for f in z.infolist() ]

    def checkoutPNG(self, page_format):
        oid = self.im.create_order()
        log.debug(f'new order id: {oid}')
        t = self.compute_total()
        b = mk_png_req(oid, self.positions, t, manifest=True)
        d = self.im.checkout_png(b)
        h = {}
        w = {}
        if d.get('link'):
            pngs = self.retrievePNGs(d['link'])
            vs = []
            for png in pngs:
                w['png_bin'] = png
                Tuple = collections.namedtuple('Png', w)
                vs.append(Tuple(**w))
            w = { 'voucher': vs }
            Tuple = collections.namedtuple('Voucher', w)
            v = Tuple(**w)
            w = { 'voucherList': v }
            Tuple = collections.namedtuple('VoucherList', w)
            v = Tuple(**w)
            h['shoppingCart'] = v
        if d.get('manifestLink'):
            h['manifest_pdf_bin'] = self.im.ses.get(d['manifestLink']).content
        Tuple = collections.namedtuple('Checkout', h)
        r = Tuple(**h)
        return r

    def retrievePreviewPDF(self, prod_code, page_format, layout = "AddressZone"):
        layout = translate_layout(layout)
        d = self.im.preview_pdf(mk_pdf_preview_req(page_format, prod_code, layout=layout))
        return d['link']

    def retrievePreviewPNG(self, prod_code, layout = "AddressZone"):
        layout = translate_layout(layout)
        d = self.im.preview_png(mk_png_preview_req(prod_code, layout=layout))
        return d['link']

    def retrievePageFormats(self):
        return self.im.get_formats()






"""Deutsche Post Internetmarke REST API client.

Deutsche Post released its `Internetmarke REST API`_ in September, 2024, as a successor to its
Internetmarke `SOAP API`_, which it deprecated in August, 2024 and announced to
turn it off at the end of 2025.

.. _SOAP API: https://developer.dhl.com/sites/default/files/2023-08/quick-Guide%20INTERNETMARKE.pdf

For sending postage requests to Deutsche Post a :class:`Session` object is required.
The body parameter that is necessary for some requests can be constructed with the
``mk_*`` functions.

For the actual checkout the total price need to be supplied and the checkout only succeeds
if your local computation matches the remote one. See also the :func:`calc_total` and :func:`compute_total` helper functions.

Example
-------

The following snippets buys a registered mail postage PDF label that can be directly printed to a C6 envelope::

    import inema.rest as ir

    im   = ir.Session('some client id', 'some client secret',
               'some portokasse user', 'some portokasse password')
    oid  = im.create_order()
    p    = ir.mk_pdf_pos(1002,
               sender=ir.mk_addr('Erika Mustermann', 'Heidestr. 17', '51147', 'Köln'),
               receiver=ir.mk_addr('Bernd Müller', 'Nußhäherstraße 10', '12345', 'München')   )
    t    = ir.calc_total(p)
    body = ir.mk_pdf_req(oid, 5, p, t)
    d    = im.checkout_pdf(body, 'postage.pdf')

Alternatively, it's also possible to check out a list of label positions, e.g.
to create multiple pages or when printing multiple labels in columns and rows to one page.


Note
----

The new `Internetmarke REST API`_ is structured quite similarly to the previous SOAP API.
On the one hand this might simplify migration, because the workflow
is basically the same, but on the other hand the REST API aguably is non-idiomatic
and contains too many surprises such that it's harder to adopt than necessary.

Hint
----
    This package contains default Deutsche Post prices, but some customers may have discounts in place.

Important
---------
   A fresh 'Portokasse' needs to allow-list the 'application' that is identified by the client id.
   This can be accomplished most easily by a first session login attempt that will fail,
   but also trigger an authorization request that can be acknowledged in the 'Portokasse' account web UI,
   after the fact.

Error
-----

This module uses the requests package for all HTTP requests.
Consequently, REST API HTTP error status codes are reported via raised
``requests.exceptions.HTTPError`` exceptions.
When caught, the remote error may be obtained like this::

    try:
        Session im(client, secret, user, password)
        // some requests, checkout, whatever ...
    except requests.exceptions.HTTPError as e:
        d = e.response.json()
        print(f'Remote errror: {d}')

Examples of such an error::

    {'statusCode': '400',
     'title': 'Bad Request',
     'description': 'positions[4].position.labelX: must be greater than or equal to 1',
     'instance': 'PCF-A1033'}

Another error example::

    {'statusCode': '400',
     'title': 'invalidTotalAmount',
     'description': 'The total amount of the order is invalid! Reason: The total cost of the shopping cart is invalid! PPL-total: 415, Cart: 1',
     'instance': 'PCF-A1031'}


.. _Internetmarke REST API:
   https://developer.dhl.com/api-reference/deutsche-post-internetmarke-post-parcel-germany

"""


# SPDX-License-Identifier: LGPL-3.0-or-later
# SPDX-FileCopyrightText: © 2025 Georg Sauthoff <mail@gms.tf>


import enum
from decimal import Decimal
import io
import requests

from .data import products as default_products



class Layout(enum.StrEnum):
    ADDRESS_ZONE  = 'ADDRESS_ZONE'
    FRANKING_ZONE = 'FRANKING_ZONE'



def mk_addr(name, line, postcode, city, country='DEU', header=None, line2=None):
    """Create address for postage label position object.

    The result of this function can be used to supply the sender and receivers parameters
    of the :func:`mk_pdf_pos` and :func:`mk_png_pos` functions.

    .. _ISO 3166-1 alpha-3:
       https://en.wikipedia.org/wiki/ISO_3166-1_alpha-3

    Parameters
    ----------
    name : str
        First name and last name, company name or something like that.
    line : str
        Street and housenumber or postbox.
        NB: It's required to be non-empty even though there are some real world postal addresses
        where only name, postcode and city are specified.
    postcode : str
        Postleitzahl
    country : str, default: 'DEU'
        `ISO 3166-1 alpha-3`_ three-letter country code.
        NB: when the DEU default is used the remote service leaves the country line blank.
    header : str, optional
        Another line above the name field, e.g. put company here and department or person in the name field then.
    line2 : str, optional
        Another line below the line field.

    Returns
    -------
    dict
        Dictionary that conforms to the inema Address schema.
    """
    d = {
        'name'         : name,
        'addressLine1' : line,
        'postalCode'   : str(postcode),
        'city'         : city,
        'country'      : country
    }
    if header is not None:
        d['additionalName'] = header
    if line2  is not None:
        d['addressLine2']   = line2
    return d


def mk_pdf_pos(product_code, column=1, row=1, page=1, layout='ADDRESS_ZONE', sender=None, receiver=None):
    """Create a postage PDF label position.

    The result (or a list of such results) can be used as a parameter of a :func:`mk_png_req` call.

    Address information is optional, but if supplied both sender and receiver need to be supplied.

    Parameters
    ----------
    product_code: int
        Product identifier from the Deutsche Post price list.
        See also :attr:`inema.data.products` or `data/products.json` or the Deutsche Post products webservice.
    column : int, default: 1
        when buying multiple positions and the format has multiple columns.
        one-based indexing
    row : int, default: 1
        when buying multiple positions and the format has multiple rows.
        one-based indexing
    page : int, default: 1
        when buying multiple positions and not all fit on one page.
        one-based indexing
    layout: { 'ADDRESS_ZONE', 'FRANKING_ZONE' }, default: 'ADDRESS_ZONE'
        use the default when you supply sender and receiver and FRANKING_ZONE when not.
        The default layout also works without sender and receiver,
        but FRANKING_ZONE is less crammed then.
    sender: dict
        Sender address information, created by :func:`mk_addr`.
    receiver: dict
        Receiver address information, created by :func:`mk_addr`.

    Returns
    -------
    dict
        Dictionary that conforms to the inema AppShoppingCartPDFPosition schema.

    Raises
    ------
    ValueError
        In case a parameter violates the schema in an obvious way.
    KeyError
        Invalid layout parameter.
    """
    layout = str(Layout[layout])
    d = {
        'positionType'  : 'AppShoppingCartPDFPosition',
        'position': {
            'labelX'    : int(column),
            'labelY'    : int(row),
            'page'      : int(page)
        },
        'productCode'   : int(product_code),
        'voucherLayout' : layout
    }
    if sender is not None:
        if receiver is None:
            raise ValueError('incomplete address')
        d['address'] = {
            'sender'   : sender,
            'receiver' : receiver
        }
    elif sender is None and receiver is not None:
        raise ValueError('incomplete address')
    if column < 1 or row < 1 or page < 1:
        raise ValueError('value violated one-based indexing')
    return d


def mk_png_pos(product_code, layout='ADDRESS_ZONE', sender=None, receiver=None):
    """Create a postage PNG label position.

    The result (or a list of such results) can be used as a parameter of a :func:`mk_png_req` call.

    Address information is optional, but if supplied both sender and receiver need to be supplied.

    Parameters
    ----------
    product_code: int
        Product identifier from the Deutsche Post price list.
        See also :attr:`inema.data.products` or `data/products.json` or the Deutsche Post products webservice.
    layout: { 'ADDRESS_ZONE', 'FRANKING_ZONE' }, default: 'ADDRESS_ZONE'
        use the default when you supply sender and receiver and FRANKING_ZONE when not.
        The default layout also works without sender and receiver,
        but FRANKING_ZONE is less crammed then.
    sender: dict
        Sender address information, created by :func:`mk_addr`.
    receiver: dict
        Receiver address information, created by :func:`mk_addr`.

    Returns
    -------
    dict
        Dictionary that conforms to the inema AppShoppingCartPosition schema.

    Raises
    ------
    ValueError
        In case a parameter violates the schema in an obvious way.
    KeyError
        Invalid layout parameter.
    """
    layout = str(Layout[layout])
    d = {
        'positionType'  : 'AppShoppingCartPosition',
        'productCode'   : product_code,
        'voucherLayout' : layout
    }
    if sender is not None:
        if receiver is None:
            raise ValueError('incomplete address')
        d['address'] = {
            'sender'   : sender,
            'receiver' : receiver
        }
    return d


def mk_pdf_req(order_id, format_id, positions, total, manifest=False):
    """
    Create PDF checkout request.

    The result can be used as body parameter of a :meth:`Session.checkout_pdf` call.

    Parameters
    ----------
    order_id : str
        Order ID returned by a previous :meth:`Session.create_order` call.
        Each request requires a fresh order ID.
    format_id: int
        Page format ID.
        See also :attr:`inema.data.formats` or `data/formats.json`
    positions: list or object
        Value or list of values created by :func:`mk_pdf_pos`.
    total: integer
        Total monetary amount of all positions. See also :func:`calc_total` and :func:`compute_total`
        for a helper.
    manifest: bool, default: False
        Also request manifest link.

    Returns
    -------
    dict
        Dictionary that conforms to the inema AppShoppingCartPDFRequest schema.

    Raises
    ------
    ValueError
        In case a parameter violates the schema in an obvious way.
    """
    if type(positions) is not list:
        positions = [ positions ]
    for p in positions:
        if p['positionType'] != 'AppShoppingCartPDFPosition':
            raise ValueError('pdf request only allows pdf positions')
    d = {
        'type'         : 'AppShoppingCartPDFRequest',
        'shopOrderId'  : str(order_id),
        'pageFormatId' : int(format_id),
        'positions'    : positions,
        'total'        : int(total)
    }
    if manifest:
        d['createManifest'] = True
    return d


def mk_png_req(order_id, positions, total, manifest=False):
    """
    Create PNG checkout request.

    The result can be used as body parameter of a :meth:`Session.checkout_png` call.

    Parameters
    ----------
    order_id : str
        Order ID returned by a previous :meth:`Session.create_order` call.
    positions: list or object
        Value or list of values created by :func:`mk_pdf_pos`.
    total: integer
        Total monetary amount of all positions. See also :func:`calc_total` and :func:`compute_total`
        for a helper.
    manifest: bool, default: False
        Also request manifest link.

    Returns
    -------
    dict
        Dictionary that conforms to the inema AppShoppingCartPNGRequest schema.

    Raises
    ------
    ValueError
        In case a parameter violates the schema in an obvious way.
    """
    if type(positions) is not list:
        positions = [ positions ]
    for p in positions:
        if p['positionType'] != 'AppShoppingCartPosition':
            raise ValueError('pdf request only allows pdf positions')
    d = {
        'type'        : 'AppShoppingCartPNGRequest',
        'shopOrderId' : str(order_id),
        'positions'   : positions,
        'total'       : int(total)
    }
    if manifest:
        d['createManifest'] = True
    return d


def mk_cancel_req(order_id=None, vouchers=None):
    """Create postage cancellation request.

    Parameters
    ----------
    order_id: str, optional
        Order ID used in a previous :meth:`Session.checkout_pdf` or :meth:`Session.checkout_png` call.
    vouchers: list, optional
        List of voucher ID prevously returned by :meth:`Session.checkout_pdf` or :meth:`Session.checkout_png` call
        or as read from the labels.
        Alternatively, list of pairs of voucher and tracking IDs.

    Returns
    -------
    dict
        Dictionary that conforms to the inema RetoureVouchersRequest schema.

    Raises
    ------
    ValueError
        In case a parameter violates the schema in an obvious way.
    """
    if order_id is None and vouchers is None:
        raise ValueError('you have to specify at least one order id or voucher')
    d = {
        'shoppingCart': { }
    }
    if order_id is not None:
        d['shoppingCart']['shopOrderId'] = str(order_id)

    def f(v):
        match v:
            case a, b:
                return { 'voucherId': a, 'trackId': b }
            case a:
                return { 'voucherId': a }

    if vouchers is not None:
        if type(vouchers) is not list:
            vouchers = [ vouchers ]
        d['shoppingCart']['voucherList'] = [ f(v) for v in vouchers ]
    if 'voucherList' in d['shoppingCart'] and len(d['shoppingCart']['voucherList']) == 0:
        raise ValueError('voucher list must not be empty')
    return d


def mk_pdf_preview_req(format_id, product_code, layout='ADDRESS_ZONE'):
    """Create PDF preview label request.

    The result can be used as body parameter to a :meth:`Session.preview_pdf` call.

    Note that the preview doesn't include any address information and thus is only
    useful for checking the placement of the QR code.

    Parameters
    ----------
    format_id : int
        Page format ID.
        See also :attr:`inema.data.formats` or `data/formats.json`
    product_code: int
        Product identifier from the Deutsche Post price list.
        See also :attr:`inema.data.products` or `data/products.json` or the Deutsche Post products webservice.
    layout: { 'ADDRESS_ZONE', 'FRANKING_ZONE' }, default: 'ADDRESS_ZONE'
        use the default when you supply sender and receiver and FRANKING_ZONE when not.
        The default layout also works without sender and receiver,
        but FRANKING_ZONE is less crammed then.

    Returns
    -------
    dict
        Dictionary that conforms to the inema AppShoppingCartPreviewPDFRequest schema.

    Raises
    ------
    KeyError
        Invalid layout parameter.
    """
    layout = str(Layout[layout])
    d = {
        'type'          : 'AppShoppingCartPreviewPDFRequest',
        'pageFormatId'  : int(format_id),
        'productCode'   : int(product_code),
        'voucherLayout' : layout
    }
    return d


def mk_png_preview_req(product_code, layout='ADDRESS_ZONE'):
    """Create PNG preview label request.

    The result can be used as body parameter to a :meth:`Session.preview_png` call.

    Note that the preview doesn't include any address information and thus is only
    useful for checking the placement of the QR code.

    Parameters
    ----------
    product_code: int
        Product identifier from the Deutsche Post price list.
        See also :attr:`inema.data.products` or `data/products.json` or the Deutsche Post products webservice.
    layout: { 'ADDRESS_ZONE', 'FRANKING_ZONE' }, default: 'ADDRESS_ZONE'
        use the default when you supply sender and receiver and FRANKING_ZONE when not.
        The default layout also works without sender and receiver,
        but FRANKING_ZONE is less crammed then.

    Returns
    -------
    dict
        Dictionary that conforms to the inema AppShoppingCartPreviewPDFRequest schema.

    Raises
    ------
    KeyError
        Invalid layout parameter.
    """
    layout = str(Layout[layout])
    d = {
        'type'          : 'AppShoppingCartPreviewPNGRequest',
        'productCode'   : int(product_code),
        'voucherLayout' : layout
    }
    return d



def check_health(api_base = 'https://api-eu.dhl.com/post/de/shipping/im/v1/'):
    """Check the availability and version of the Deutsche Post INTERNETMARKE REST API endpoint.

    NB: doesn't require a session login.

    Example
    -------
    >>> check_health()
    {'name'    : 'pp-post-internetmarke',
     'version' : 'v1.1.18',
     'rev'     : '35',
     'env'     : 'prod-eu'}
    """
    r = requests.get(api_base)
    r.raise_for_status()
    d = r.json()
    return d['amp']



class Session:
    """Internetmarke REST API Connection class.

    An object of this class primarily wraps the requests http session (`ses`),
    the bearer token obtained after authorization (`token`) and the current
    balance of the user's Portokasse wallet (`balance`).

    The constructor authorises to Internetmarke REST API endpoint.

    Parameters
    ----------
    client : str
        Client ID, a.k.a. the API key credential of your 'Post DE
        Internetmarke (Post & Parcel Germany)' API application that is
        registered in your developer.dhl.com user account.
    secret : str
        Client secret, a.k.a. the API Secret credential of yoru 'Post DE
        Internetmarke (Post & Parcel Germany)' API application that is
        registered in your developer.dhl.com user account.
    user: str
        username of your portokasse.deutschepost.de Portokasse account
    password: str
        password of your portokasse.deutschepost.de Portokasse account

    Raises
    ------
    requests.exceptions.HTTPError
        In case API endpoint responds with an error.

    Example
    -------
    >>> im = Session im(client, secret, user, password)
    >>> im.auth_resp
    {'userToken'              : 'seCreTbeArer=',
     'access_token'           : 'seCreTbeArer=',
     'walletBalance'          : 3285,
     'showTermsAndConditions' : False,
     'infoMessage'            : '',
     'issued_at'              : 'Sun, 14 Sep 2025 14:15:18 GMT',
     'expires_in'             : 86400,
     'token_type'             : 'BearerToken',
     'external_customer_id'   : 'juser-0815',
     'authenticated_user'     : 'myportokasse@example.org'}
    >>> im.balance
    3285
    >>> im.api_base
    'https://api-eu.dhl.com/post/de/shipping/im/v1/'
    >>> type(im.ses)
    requests.sessions.Session
    """


    def __init__(self, client, secret, user, password, api_base = 'https://api-eu.dhl.com/post/de/shipping/im/v1/'):
        ses = requests.Session()
        ses.headers.update({'User-Agent': 'python-requests/python-inema'})
        r = ses.post(api_base + 'user',
                 data={
                     'grant_type'   : 'client_credentials',
                     'client_id'    : client,
                     'client_secret': secret,
                     'username'     : user,
                     'password'     : password,
                     }
                )
        r.raise_for_status()
        d     = r.json()
        token = d['access_token']
        ses.headers.update({ 'Authorization': f'Bearer {token}' })
        self.api_base  = api_base
        self.ses       = ses
        self.auth_resp = d
        self.balance   = d['walletBalance']


    def profile(self):
        """Request your Portokasse profile fields.

        Example
        -------
        >>> im.profile()
        {'ekp'              : None,
         'company'          : 'ACME Inc',
         'title'            : 'Dr.',
         'invoiceType'      : 'PAPER',
         'invoiceFrequency' : 'DECADE',
         'mail'             : 'myportokasse@example.org',
         'firstname'        : 'Erika',
         'lastname'         : 'Mustermann',
         'street'           : 'Heidestr.',
         'houseNo'          : '17',
         'zip'              : '51147',
         'city'             : 'Köln',
         'country'          : 'DEU',
         'phone'            : None,
         'pobox'            : None,
         'poboxZip'         : None,
         'poboxCity'        : None}

        Raises
        ------
        requests.exceptions.HTTPError
            In case API endpoint responds with an error.
        """
        r = self.ses.get(self.api_base + 'user/profile')
        r.raise_for_status()
        d = r.json()
        return d


    def create_order(self):
        """Initialize new order.

        Returns new order ID that is required e.g. for building the body
        of :meth:`checkout_pdf` or :meth:`checkout_png`.

        Returns
        -------
        str
            New order ID

        Raises
        ------
        requests.exceptions.HTTPError
            In case API endpoint responds with an error.
        """
        r = self.ses.post(self.api_base + 'app/shoppingcart')
        r.raise_for_status()
        d = r.json()
        order_id = d['shopOrderId']
        return order_id


    def list_order(self, order_id):
        """Request a checked out order.

        Parameters
        ----------
        order_id: str
            Order ID used in a previous :meth:`checkout_pdf` or :meth:`checkout_png` call.

        Raises
        ------
        requests.exceptions.HTTPError
            In case API endpoint responds with an error.
            Example: When order ID wasn't used to checkout something, yet.

        Example
        -------
        >>> im.list_order(123456)
        {'type'           : 'CheckoutShoppingCartAppResponse',
         'link'           : 'https: //internetmarke.deutschepost.de/PcfExtensionWeb/document?keyphase=1&data=blAHblAh%3D',
         'manifestLink'   : None,
         'shoppingCart'   : {'shopOrderId': '123456',
          'voucherList'   : [{'voucherId': 'A00002300A0000000815', 'trackId': None}]},
         'walletBallance' : None}
        """
        r = self.ses.get(self.api_base + f'app/shoppingcart/{order_id}')
        r.raise_for_status()
        d = r.json()
        return d


    def download(self, src, filename):
        """Download convenience function.

        Downloads through the wrapped requests session,
        while not wasting too much memory in case the downloaded file is very large.

        Parameters
        ----------
        src: str
            source URL
        filename: str or path_like
            target filename

        Raises
        ------
        requests.exceptions.HTTPError
            In case API endpoint responds with an error.
        """
        with self.ses.get(src, stream=True) as x:
            x.raise_for_status()
            with open(filename, 'wb') as f:
                for bs in x.iter_content(chunk_size=128 * 1024):
                    f.write(bs)


    def checkout_pdf(self, body, filename=None):
        """Buy PDF labels.

        Parameters
        ----------
        body: dict
            Request body, usually created with :func:`mk_pdf_req`.
        filename: str or path_like, optional
            Filename for convenience.
            If not specified, the PDF can alternatively be retrieved from the URL
            returned in the 'link' field, e.g. using the requests session available
            via the `ses` object property.

        Returns
        -------
        dict
            Response fields

        Raises
        ------
        requests.exceptions.HTTPError
            In case API endpoint responds with an error.

        Example
        -------
        >>> im.checkout_pdf(body_with_4_positions, 'labels.pdf')
        {'type': 'CheckoutShoppingCartAppResponse',
         'link': 'https://internetmarke.deutschepost.de/PcfExtensionWeb/document?keyphase=1&data=rANdOMdaTaString%3D',
         'manifestLink': None,
         'shoppingCart': {'shopOrderId': '123456789',
             'voucherList': [{'voucherId': 'A00002300A0000000123', 'trackId': None},
                 {'voucherId': 'A00002300A0000000122', 'trackId': None},
                 {'voucherId': 'A00002300A0000000121', 'trackId': None},
                 {'voucherId': 'A00002300A0000000120', 'trackId': None}]},
         'walletBallance': 4205}
        """
        r = self.ses.post(self.api_base + 'app/shoppingcart/pdf', json=body)
        r.raise_for_status()
        d = r.json()
        if 'walletBallance' in d:
            self.balance = d['walletBallance']
        if filename is not None:
            self.download(d['link'], filename)
        return d


    def checkout_png(self, body, filename=None):
        """Buy PNG labels.

        Parameters
        ----------
        body: dict
            Request body, usually created with :func:`mk_png_req`.
        filename: str or path_like, optional
            Filename for convenience.
            NB: The Internetmarke API endpoint always returns PNG labels as ZIP archive,
            i.e. even when just requesting a single label.
            If not specified, the ZIP result can alternatively be retrieved from the URL
            returned in the 'link' field, e.g. using the requests session available
            via the `ses` object property.

        Returns
        -------
        dict
            Response fields

        Raises
        ------
        requests.exceptions.HTTPError
            In case API endpoint responds with an error.

        """
        r = self.ses.post(self.api_base + 'app/shoppingcart/png', json=body)
        r.raise_for_status()
        d = r.json()
        if 'walletBallance' in d:
            self.balance = d['walletBallance']
        if filename is not None:
            self.download(d['link'], filename)
        return d


    def cancel(self, body):
        """Request cancellation of previously checked out postage.

        Parameters
        ----------
        body: dict
            Request body, usually created with :func:`mk_cancel_req`.

        Returns
        -------
        dict
            Response fields

        Raises
        ------
        requests.exceptions.HTTPError
            In case API endpoint responds with an error.

        Example
        -------
        >>> im.cancel(mk_cancel_req(oid))
        { 'shopRetoureId': '123456789', 'retoureTransactionId': '23230815' }
        """
        r = self.ses.post(self.api_base + f'app/retoure', json=body)
        r.raise_for_status()
        d = r.json()
        return d


    def list_cancel(self, retoure_id=None, transaction_id=None, start=None, end=None):
        """Request status of previously requested cancellations.

        Parameters
        ----------
        retoure_id: str, optional
            shopRetoureId field previously returned by a :meth:`cancel` call
        transaction_id: int, optional
            retoureTransactionId field previously returned by a :meth:`cancel` call
        start: str, optional
            start of query period in ISO 8601 datetime format with ``T`` delimiter and without timezone
        end: str, optional
            end of query period in ISO 8601 datetime format with ``T`` delimiter and without timezone

        Returns
        -------
        dict
            Response fields

        Raises
        ------
        requests.exceptions.HTTPError
            In case API endpoint responds with an error.

        Example
        -------
        >>> im.list_cancel(start='2025-09-13T23:45:00')
        [{'retoureTransactionId' : 1234,
          'shopRetoureId'        : '5678',
          'totalCount'           : 1,
          'countStillOpen'       : 0,
          'retourePrice'         : 415,
          'creationDate'         : '14092025-143423',
          'serialnumber'         : 'A00230007A',
          'refundedVouchers'     : [{'voucherId': 'A00230007A000000042C', 'trackId': None}],
          'notRefundedVouchers'  : []},
         {'retoureTransactionId' : 1233,
          'shopRetoureId'        : '5677',
          'totalCount'           : 1,
          'countStillOpen'       : 0,
          'retourePrice'         : 215,
          'creationDate'         : '14092025-142407',
          'serialnumber'         : 'A00230009A',
          'refundedVouchers'     : [{'voucherId': 'A00230009A000000023F', 'trackId': None}],
          'notRefundedVouchers'  : []}]
        """
        ps = {}
        if retoure_id is not None:
            ps['shopRetoureId'] = str(retoure_id)
        if transaction_id is not None:
            ps['retoureTransactionId'] = int(transaction_id)
        if start is not None:
            ps['startDate'] = start
        if end is not None:
            ps['endDate'] = end
        r = self.ses.get(self.api_base + f'app/retoure', params=ps)
        r.raise_for_status()
        d = r.json()
        d = d['RetrieveRetoureStateResponse']
        return d


    def preview_pdf(self, body, filename=None):
        """Request preview PDF.

        Note that Deutsche Post doesn't support previews with addresses.
        Thus, such previews are only got for testing the QR code placement.

        Parameters
        ----------
        body: dict
            Request body, usually created with :func:`mk_pdf_preview_req`
        filename: str or path_like, optional
            Filename for convenience.
            If not specified, the PDF can alternatively be retrieved from the URL
            returned in the 'link' field, e.g. using the requests session available
            via the `ses` object property.

        Returns
        -------
        dict
            Response fields

        Raises
        ------
        requests.exceptions.HTTPError
            In case API endpoint responds with an error.

        Example
        -------
        >>> im.preview_pdf(mk_pdf_preview_req(5, 1), 'letter.pdf')
        {'link': 'https://internetmarke.deutschepost.de/PcfExtensionWeb/preview?keyphase=1&data=ranDodmsTringDaTA'}
        """
        r = self.ses.post(self.api_base + 'app/shoppingcart/pdf', params={ 'validate': True }, json=body)
        r.raise_for_status()
        d = r.json()
        if filename is not None:
            self.download(d['link'], filename)
        return d


    def preview_png(self, body, filename=None):
        """Request preview PNG.

        Note that Deutsche Post doesn't support previews with addresses.
        Thus, such previews are only got for testing the QR code placement.

        NB: In contrast to :meth:`checkout_png` the API endpoint returns
        the preview directly as PNG and _not_ inside a ZIP archive.

        Parameters
        ----------
        body: dict
            Request body, usually created with :func:`mk_png_preview_req`
        filename: str or path_like, optional
            Filename for convenience.
            If not specified, the PNG can alternatively be retrieved from the URL
            returned in the 'link' field, e.g. using the requests session available
            via the `ses` object property.

        Returns
        -------
        dict
            Response fields

        Raises
        ------
        requests.exceptions.HTTPError
            In case API endpoint responds with an error.
        """
        r = self.ses.post(self.api_base + 'app/shoppingcart/png', params={ 'validate': True }, json=body)
        r.raise_for_status()
        d = r.json()
        if filename is not None:
            self.download(d['link'], filename)
        return d


    def charge(self, cent):
        """Top up the Portokasse.

        Parameters
        ----------
        cent: int
            monetary amount to transfer to your 'Portokasse' in euro cent units

        Returns
        -------
        dict
            Response fields

        Raises
        ------
        requests.exceptions.HTTPError
            In case API endpoint responds with an error.

        Example
        -------
        Top up the wallet (Portokasse) with 7 €, at a balance of 85.05 €:

        >>> im.charge(700)
        {'shopOrderId': '123456', 'walletBalance': 9205}
        >>> im.balance
        9205

        """
        r = self.ses.put(self.api_base + 'app/wallet', params={ 'amount': int(cent) })
        r.raise_for_status()
        d = r.json()
        self.balance = d['walletBalance']
        return d


    def get_formats(self):
        """Query format catalogue.

        See Also
        --------
        :attr:`inema.data.formats` : or `data/formats.json` for a cached version of that data.
        :func:`formats2json` : helper to write the result as json
        :func:`formats2json` : write result in a more compact and normalized CSV format

        See also :func:`

        Returns
        -------
        list
            List of format dictionaries.

        Example
        -------
        >>> xs = im.get_formats()
        >>> xs[0]
        {'id'                : 1,
         'isAddressPossible' : True,
         'isImagePossible'   : False,
         'name'              : 'DIN A4 Normalpapier',
         'description'       : None,
         'pageType'          : 'REGULARPAGE',
         'pageLayout'        : {'size': {'x': 210, 'y': 297},
             'orientation'   : 'PORTRAIT',
             'labelSpacing'  : {'x': 0, 'y': 0},
             'labelCount'    : {'labelX': 2, 'labelY': 5},
             'margin'        : {'top': 31, 'bottom': 31, 'left': 15, 'right': 15}}}
        """
        r = self.ses.get(self.api_base + 'app/catalog', params={ 'types': 'PAGE_FORMATS'} )
        r.raise_for_status()
        def f(x):
            if x.endswith('.0'):
                return int(x[:-2])
            else:
                raise ValueError('unexpected real float')
        d = r.json(parse_float=f)
        fs = d['pageFormats']
        fs.sort(key=lambda f : f['id'])
        return fs



def compute_total(positions, im, pid2price=None):
    """Compute total of a positions based on prices available online.

    NB: This function requires and imports the pdfminer package.

    Parameters
    ----------
    positions : list or dict or int
       Single product code or list of them, either as raw int or dictonary with ``productCode`` key.
    im : Session
       :class:`Session` object for querying the prices.
    pid2price : dict, optional
       Dictionary to cache product prices for the current and future invocations.

    Returns
    -------
    int
        The monetary sum of all positions in euro cent.
    """
    import pdfminer.high_level
    if type(positions) is not list:
        positions = [ positions ]
    if pid2price is None:
        h = {}
    else:
        h = pid2price
    total = 0
    for p in positions:
        if type(p) is dict:
            pc = p['productCode']
        else:
            pc = p
        if  pc in h:
            total += h[pc]
            continue
        d = im.preview_pdf(mk_pdf_preview_req(5, pc))
        r = im.ses.get(d['link'])
        r.raise_for_status()
        f = io.BytesIO(r.content)
        s = pdfminer.high_level.extract_text(f)
        l = next(l for l in s.splitlines() if l.startswith('IM'))
        xs = l.split()
        t = xs[-1]
        if ',' not in t:
                raise ValueError('cannot find comma')
        t = t.replace(',', '.')
        c = Decimal(t) * Decimal('100')
        total += int(c)
        h[pc]  = int(c)
    return total


def calc_total(positions, products=default_products):
    """Compute total of a positions based on a price dictionary.

    Parameters
    ----------
    positions : list or dict or int
       Single product code or list of them, either as raw int or dictonary with ``productCode`` key.
    products : dict
       Dictionary that maps product code strings to dictionaries that contain the price at the `cost_price` key.
       See :attr:`data.products` for the default prices.

    Returns
    -------
    int
        The monetary sum of all positions in euro cent.
    """
    if type(positions) is not list:
        positions = [ positions ]
    total = 0
    for p in positions:
        if type(p) is dict:
            pc = p['productCode']
        else:
            pc = p
        t = products[str(pc)]['cost_price']
        c = Decimal(t) * Decimal('100')
        total += int(c)
    return total


def formats2json(formats, filename):
    import json
    with open(filename, 'w') as f:
        json.dump(formats, f, ensure_ascii=False, indent=2)


def formats2csv(formats, filename):
    import pandas as pd
    df = pd.json_normalize(fs)
    df.set_index('id', inplace=True)
    df.rename(columns={
        'isAddressPossible'            : 'addr_space',
        'isImagePossible'              : 'img_space',
        'pageType'                     : 'ptype',
        'pageLayout.size.x'            : 'width',
        'pageLayout.size.y'            : 'height',
        'pageLayout.orientation'       : 'orientation',
        'pageLayout.labelSpacing.x'    : 'hpadding',
        'pageLayout.labelSpacing.y'    : 'vpadding',
        'pageLayout.labelCount.labelX' : 'hcount',
        'pageLayout.labelCount.labelY' : 'vcount',
        'pageLayout.margin.top'        : 'tmargin',
        'pageLayout.margin.bottom'     : 'bmargin',
        'pageLayout.margin.left'       : 'lmargin',
        'pageLayout.margin.right'      : 'rmargin',
    }, inplace=True)
    df.to_csv(filename)





#!/usr/bin/python
# -*- coding: utf-8 -*-
# SPDX-Identifier: LGPL-3.0-or-later

from datetime import date
import json
from lxml import etree
from zeep import Client
from zeep.wsse.username import UsernameToken
from pkg_resources import resource_stream, resource_listdir
import pkg_resources
import requests, zipfile
import io
import logging
import decimal

from .utils import compute_1c4a_hash, gen_timestamp

__version__ = pkg_resources.get_distribution(__package__).version

_logger = logging.getLogger(__name__)


# Read the most recent inema/data/products-YYYY-MM-DD.json where YYYY-MM-DD is
# not more recent than today. Fall back to inema/data/products.json.
# This allows to ship multiple products.json files for announced future price/product
# changes.
def load_products():
    ps = [ e for e in resource_listdir(__name__, 'data')
            if e.startswith('products-') and e.endswith('.json')
            and e <= 'products-{}.json'.format(date.today().isoformat()) ]
    ps.sort()
    ps.insert(0, 'products.json')
    products_json = resource_stream(__name__, 'data/' + ps[-1]).read().decode("utf-8")
    products = json.loads(products_json)
    return products

default_products = load_products()

formats_json = resource_stream(__name__, "data/formats.json").read().decode("utf-8")
formats = json.loads(formats_json)

class ProductInformation(object):
    """Product Webservice ProdWS SOAP API client.

    The `ProdWS`_ SOAP API is provided by Deutsche Post for getting a list of products
    and their prices.

    .. _ProdWS: https://developer.dhl.com/api-reference/deutsche-post-internetmarke-post-parcel-germany?language_content_entity=de#additional-information-section/informationen-zum-einbinden-des-produkt-webservice-_28prodws_29

    Note
    ----

    As of 2025 the Internetnetmarke REST API does NOT offer equivalent functionality.
    """

    wsdl_url = 'https://prodws.deutschepost.de:8443/ProdWSProvider_1_1/prodws?wsdl'

    def __init__(self, username, password, mandantid):
        self.client = Client(self.wsdl_url, wsse=UsernameToken(username, password))
        self.username = username
        self.password = password
        self.mandantid = mandantid

    def getProductList(self):
        s = self.client.service
        r = s.getProductList(mandantID=self.mandantid, dedicatedProducts=True, responseMode=0)
        _logger.info("getProductList result: %s", r)

        return r


class Internetmarke(object):
    wsdl_url = 'https://internetmarke.deutschepost.de/OneClickForAppV3/OneClickForAppServiceV3?wsdl'

    # generate a 1C4A SOAP header
    def gen_1c4a_hdr(self, partner_id, key_phase, key):
        nsmap={'soapenv': 'http://schemas.xmlsoap.org/soap/envelope/',
               'v3':'http://oneclickforpartner.dpag.de'}
        r = etree.Element("{http://schemas.xmlsoap.org/soap/envelope/}Header", nsmap = nsmap)
        p = etree.SubElement(r, "{http://oneclickforpartner.dpag.de}PARTNER_ID")
        p.text = partner_id
        t = etree.SubElement(r, "{http://oneclickforpartner.dpag.de}REQUEST_TIMESTAMP")
        t.text = gen_timestamp()
        k = etree.SubElement(r, "{http://oneclickforpartner.dpag.de}KEY_PHASE")
        k.text = key_phase
        s = etree.SubElement(r, "{http://oneclickforpartner.dpag.de}PARTNER_SIGNATURE")
        s.text = compute_1c4a_hash(partner_id, t.text, key_phase, key)
        return [p, t, k, s]

    def get_product_price_by_id(self, ext_prod_id):
        price_str = self.products[str(ext_prod_id)]['cost_price']
        return int(decimal.Decimal(price_str) * 100)

    def __init__(self, partner_id, key, key_phase="1", products=None):
        self.client = Client(self.wsdl_url)
        self.partner_id = partner_id
        self.key_phase = key_phase
        self.key = key
        self.soapheader = self.gen_1c4a_hdr(self.partner_id, self.key_phase, self.key)
        self.positions = []
        if products is None:
            self.products = default_products
        else:
            self.products = products


    def authenticate(self, username, password):
        s = self.client.service
        r = s.authenticateUser(_soapheaders= self.soapheader, username=username, password=password)
        self.user_token = r.userToken
        self.wallet_balance = r.walletBalance

    def retrievePNGs(self, link):
        _logger.info("Retrieving PNGs from %s", link)
        r = requests.get(link, stream=True)
        z = zipfile.ZipFile(io.BytesIO(r.content))
        return [ z.read(f.filename) for f in z.infolist() ]

    def retrieve_manifest(self, link):
        _logger.info("Retrieving Manifest from %s", link)
        r = requests.get(link, stream=True)
        return r.content

    def retrievePreviewPNG(self, prod_code, layout = "AddressZone"):
        s = self.client.service
        r = s.retrievePreviewVoucherPNG(_soapheaders = self.soapheader,
                                        productCode = prod_code,
                                        voucherLayout = layout)
        _logger.info("retrievePreviewPNG result: %s", r)
        return r

    def retrievePreviewPDF(self, prod_code, page_format,
        layout = "AddressZone"):
        s = self.client.service
        r = s.retrievePreviewVoucherPDF(_soapheaders = self.soapheader,
                                        productCode = prod_code,
                                        pageFormatId = page_format,
                                        voucherLayout = layout)
        _logger.info("retrievePreviewPPDF result: %s", r)
        return r

    def retrievePageFormats(self):
        s = self.client.service
        r = s.retrievePageFormats(_soapheaders = self.soapheader)
        _logger.info("retrievePageFormats result: %s", r)
        return r


    def add_position(self, position):
        _logger.info("Adding position to basket: %s", position)
        self.positions.append(position)

    def clear_positions(self):
        _logger.info("Clearing positions from basket")
        self.positions = []

    def compute_total(self):
        total = 0
        for p in self.positions:
            total += self.get_product_price_by_id(p.productCode)
        return total

    def checkoutPDF(self, page_format):
        s = self.client.service
        # FIXME: convert ShoppingCartPosition to ShoppingCartPDFPosition
        _logger.info("Submitting basket with %u positions, total=%f", len(self.positions), self.compute_total())
        r = s.checkoutShoppingCartPDF(_soapheaders = self.soapheader,
                                  userToken = self.user_token,
                                  pageFormatId = page_format,
                                  positions = self.positions,
                                  total = self.compute_total(),
                                  createManifest = True,
                                  createShippingList = 2)
        _logger.info("PDF checkout result: %s", r)
        if r.link:
            pdf = requests.get(r.link, stream=True)
            r.pdf_bin = pdf.content
        if r.manifestLink:
            r.manifest_pdf_bin = self.retrieve_manifest(r.manifestLink)
        return r

    def checkoutPNG(self):
        s = self.client.service
        _logger.info("Submitting basket with %u positions, total=%f", len(self.positions), self.compute_total())
        r = s.checkoutShoppingCartPNG(_soapheaders = self.soapheader,
                                  userToken = self.user_token,
                                  positions = self.positions,
                                  total = self.compute_total(),
                                  createManifest = True,
                                  createShippingList = 2)
        _logger.info("PNG checkout result: %s", r)
        if r.link:
            # retrieve PNG images and store them in result object
            pngs = self.retrievePNGs(r.link)
            for i in range(0, len(pngs)):
                r.shoppingCart.voucherList.voucher[i].png_bin = pngs[i]
        if r.manifestLink:
            # retrieve manifest PDF and store in result object
            r.manifest_pdf_bin = self.retrieve_manifest(r.manifestLink)
        return r


    def build_addr(self, street, house, zipcode, city, country, additional = None):
        zclient = self.client
        atype = zclient.get_type('{http://oneclickforapp.dpag.de/V3}Address')
        return atype(additional = additional, street = street,
                     houseNo = house, zip = zipcode, city = city,
                     country= country)

    def build_pers_name(self, first, last, salutation=None, title=None):
        zclient = self.client
        pntype = zclient.get_type('{http://oneclickforapp.dpag.de/V3}PersonName')
        pn = pntype(firstname = first, lastname = last,
                    salutation = salutation, title = title)
        return pn

    def build_comp_addr(self, company, address, person = None):
        zclient = self.client
        cntype = zclient.get_type('{http://oneclickforapp.dpag.de/V3}CompanyName')
        cn = cntype(company = company, personName = person)
        ntype = zclient.get_type('{http://oneclickforapp.dpag.de/V3}Name')
        name = ntype(companyName = cn)
        atype = zclient.get_type('{http://oneclickforapp.dpag.de/V3}NamedAddress')
        return atype(name = name, address = address)

    def build_pers_addr(self, first, last, address, salutation = None, title = None):
        zclient = self.client
        pn = self.build_pers_name(first, last, salutation, title)
        ntype = zclient.get_type('{http://oneclickforapp.dpag.de/V3}Name')
        name = ntype(personName = pn)
        atype = zclient.get_type('{http://oneclickforapp.dpag.de/V3}NamedAddress')
        return atype(name = name, address = address)

    def build_position(self, product, sender=None, receiver=None, layout="AddressZone", pdf=False, x=1, y=1, page=1):
        zclient = self.client
        abtype = zclient.get_type('{http://oneclickforapp.dpag.de/V3}AddressBinding')
        postype = zclient.get_type('{http://oneclickforapp.dpag.de/V3}VoucherPosition')
        if pdf:
            pos = 'ShoppingCartPDFPosition'
            args = { 'position': postype(labelX=x, labelY=y, page=page) }
        else:
            pos = 'ShoppingCartPosition'
            args = {}
        ptype = zclient.get_type('{http://oneclickforapp.dpag.de/V3}' + pos)
        ab = abtype(sender = sender, receiver = receiver) if sender and receiver else None
        return ptype(productCode = product, address = ab,
                voucherLayout = layout, **args)



# SPDX-Identifier: LGPL-3.0-or-later

import hashlib
from datetime import datetime
from pytz import timezone

def compute_1c4a_hash(partner_id, req_ts, key_phase, key):
    """Compute 1C4A request hash according to Section 4 of service description."""
    # trim leading and trailing spaces of each argument
    partner_id = partner_id.strip()
    req_ts = req_ts.strip()
    key_phase = key_phase.strip()
    key = key.strip()
    # concatenate with "::" separator
    inp = "%s::%s::%s::%s" % (partner_id, req_ts, key_phase, key)
    # compute MD5 hash as 32 hex nibbles
    md5_hex = hashlib.md5(inp.encode('utf8')).hexdigest()
    # return the first 8 characters
    return md5_hex[:8]

def gen_timestamp():
    """Compute a timestamp as used in the 1C4A and WaPoInt APIs."""
    de_zone = timezone("Europe/Berlin")
    de_time = datetime.now(de_zone)
    return de_time.strftime("%d%m%Y-%H%M%S")



# -*- coding: utf-8 -*-

# Implementation of the "Warenpost International ReST-API"
#
# This is a completely different interface that Deutsche Post came up many years
# after the "1C4A" Internetmarke API.  For some strange reason they didn't
# extend the old Internetmarke API to add support for the harmonized label and
# the electronic customs declaration.  Instead, they decided to implement a
# completely different API with different standards (REST vs. SOAP) with
# literally nothing in common to the old Internetmarke API

# SPDX-Identifier: LGPL-3.0-or-later

import json
import logging

import requests
from lxml import etree

from .utils import compute_1c4a_hash, gen_timestamp

_logger = logging.getLogger(__name__)

class WarenpostInt(object):
    """Represents the Warenpost Internatoinal REST interface.

    Parameters
    ----------
    key : str
        same as SCHLUESSEL_DPWN_PARTNER?
    ekp : str
        'Einheitliche Kunden- und Produktnummer' aka Geschäftskundennummer aka business customer number
    """
    def __init__(self, partner_id, key, ekp, pk_email, pk_passwd, key_phase="1", sandbox = False):
        self.sandbox = sandbox
        self.partner_id = 'DP_LT' if sandbox else partner_id
        self.key = key
        self.ekp = ekp
        self.key_phase = '1' if sandbox else key_phase
        self.pk_email = pk_email
        self.pk_passwd = pk_passwd
        if sandbox:
            self.auth_url = 'https://api-qa.deutschepost.com/v1'
            self.url = 'https://api-qa.deutschepost.com/dpi/shipping/v1'
        else:
            self.auth_url = 'https://api.deutschepost.com/v1'
            self.url = 'https://api.deutschepost.com/dpi/shipping/v1'
        self.user_token = None
        self.wallet_balance = None

    def gen_headers(self):
        """Generate the HTTP headers required for the API."""
        ret = {
            'KEY_PHASE': self.key_phase,
            'PARTNER_ID': self.partner_id,
            'Authorization': 'Bearer %s' % self.user_token
            }
        if self.sandbox:
            ret['REQUEST_TIMESTAMP'] = '16082018-122210'
            ret['PARTNER_SIGNATURE'] = '9d7c35be'
        else:
            timestamp = gen_timestamp()
            sig = compute_1c4a_hash(self.partner_id, timestamp, self.key_phase, self.key)
            ret['REQUEST_TIMESTAMP'] = timestamp
            ret['PARTNER_SIGNATURE'] = sig
        return ret

    def get_token(self):
        """Get an Access Token for further API requests."""
        url = "%s/%s" % (self.auth_url, 'auth/accesstoken')
        auth = requests.auth.HTTPBasicAuth(self.pk_email, self.pk_passwd)
        ret = requests.request('GET', url, headers=self.gen_headers(), auth=auth)
        et = etree.XML(ret.content)
        e_user_token = et.find(".//{http://oneclickforapp.dpag.de/V3}userToken")
        e_wallet_balance = et.find(".//{http://oneclickforapp.dpag.de/V3}walletBalance")
        # update status + return token
        self.user_token = e_user_token.text
        self.wallet_balance = e_wallet_balance.text
        _logger.debug("User Token: %s", self.user_token)
        _logger.info("Wallet balance: %s", self.wallet_balance)
        return e_user_token.text

    def request(self, method, suffix, json=None, headers=None):
        """Wrapper for issuing HTTP requests against the API.
           This internally generates all required headers, including Authorization."""
        url = "%s/%s" % (self.url, suffix)
        # FIXME: automatically ensure we have a [current] user_token
        h = headers.copy() if headers else {}
        h.update(self.gen_headers())
        _logger.debug("HTTP Request: %s %s: HDR: %s JSON: %s", method, url, h, json)
        r = requests.request(method, url, json=json, headers=h)
        _logger.debug("HTTP Response: %s", r.content)
        return r

    class Address(object):
        """Common Representation of a postal address.  In their infinite cluelessness,
           the developes of the Warenpost International API decided it's a good idea to
           use a flat, non-hierarchical structure with different names of fields for
           sender and recipient."""
        def __init__(self, name, addr_lines, city, country_code, postal_code='', state=None,
                     phone=None, fax=None, email=None):
            if len(name) > 30:
                raise ValueError('Maximum length of name is 30 chars')
            if len(addr_lines) > 3:
                raise ValueError('Maximum number of 3 Address Lines supported')
            if len(city) > 30:
                raise ValueError('Maximum length of city is 30 chars')
            if len(country_code) != 2:
                raise ValueError('Country must be 2-digit ISO-3166-1 code')
            if state and len(state) > 20:
                raise ValueError('Maximum length of state is 20 chars')
            if phone and len(phone) > 15:
                raise ValueError('Maximum length of phone number is 15 chars')
            if fax and len(fax) > 15:
                raise ValueError('Maximum length of fax number is 15 chars')
            if email and len(email) > 50:
                raise ValueError('Maximum length of email address is 50 chars')
            for l in addr_lines:
                if len(l) > 40:
                    raise ValueError('Maximum length of address lines is 40 chars')
            self.name = name
            self.addr_lines = addr_lines
            self.city = city
            self.postal_code = postal_code
            self.country_code = country_code
            self.state = state
            self.phone = phone
            self.fax = fax
            self.email = email

        def as_sender(self):
            """Represent an Address object as JSON fields of a sender."""
            if len(self.addr_lines) > 2:
                raise ValueError('Maximum number of 2 Address Lines supported')
            for l in self.addr_lines:
                if len(l) > 30:
                    raise ValueError('Maximum length of address lines is 30 chars')
            ret = {
                'senderName': self.name,
                'senderAddressLine1': self.addr_lines[0],
                'senderAddressLine2': self.addr_lines[1] if len(self.addr_lines) > 1 else '',
                'senderCity': self.city,
                'senderPostalCode': self.postal_code,
                'senderCountry': self.country_code,
                }
            if self.phone:
                ret['senderPhone'] = self.phone
            if self.email:
                ret['senderEmail'] = self.email
            return ret

        def as_recipient(self):
            """Represent an Address object as JSON fields of a sender."""
            ret = {
                'recipient': self.name,
                'addressLine1': self.addr_lines[0],
                'city': self.city,
                'postalCode': self.postal_code,
                'destinationCountry': self.country_code,
                }
            if len(self.addr_lines) > 1:
                ret['addressLine2'] = self.addr_lines[1]
                if len(self.addr_lines) > 2:
                    ret['addressLine3'] = self.addr_lines[2]
            if self.state:
                ret['state'] = self.state
            if self.phone:
                ret['recipientPhone'] = self.phone
            if self.fax:
                ret['recipientFax'] = self.fax
            if self.email:
                ret['recipientEmail'] = self.email
            return ret

    def build_content_item(self, line_weight_g, line_value, qty, hs_code=None, origin_cc=None,
                           desc=None):
        """Build an 'content item' in the language of the WaPoInt API. Represents one
           line on the customs form."""
        line_weight_g = int(line_weight_g)
        if line_weight_g > 2000:
            raise ValueError('Maximum line weight is 2000g')
        qty = int(qty)
        if qty > 99 or qty < 1:
            raise ValueError('Maximum line quantity is 99')
        if desc and len(desc) > 33:
            raise ValueError('Maximum length of contentPieceDescription is 33 chars')
        if hs_code and (len(hs_code) < 4 or len(hs_code) > 10):
            raise ValueError('HS-Code must be between 4 and 10 characters long')
        ret = {
            'contentPieceNetweight': line_weight_g,
            'contentPieceValue': "%.2f" % (line_value),
            'contentPieceAmount': qty,
            }
        if hs_code:
            ret['contentPieceHsCode'] = str(hs_code)
        if desc:
            ret['contentPieceDescription'] = str(desc)
        if origin_cc:
            ret['contentPieceOrigin'] = origin_cc
        return ret

    def shrink_contents_if_needed(self, contents):
        """Attempt to shrink the number of content lines below the permitted 5. We
           intentionally ignore the country of origin and group all lines by the
           HTS code. All lines sharing the same HTS code are merged to one,
           hopefully this is sufficient to get the count below 5."""
        if len(contents) < 5:
            return contents
        # group items by HTS code
        by_hts = {}
        for c in contents:
            hts = c['contentPieceHsCode'].strip()
            if hts in by_hts:
                by_hts[hts].append(c)
            else:
                by_hts[hts] = [c]
        if len(by_hts.keys()) > 5:
            raise ValueError('More than 5 distinct HTS numbers(%u); cannot merge' % (len(by_hts.keys())))
        out = []
        # generate one aggregate item per HTS code
        for k in by_hts:
            total_grams = 0
            total_value = 0
            total_qty = 0
            for c in by_hts[k]:
                total_grams += c['contentPieceNetweight']
                total_value += float(c['contentPieceValue'])
                total_qty += c['contentPieceAmount']
            aggregate = by_hts[k][0]
            aggregate['contentPieceAmount'] = total_qty
            aggregate['contentPieceValue'] = "%.2f" % (total_value)
            aggregate['contentPieceNetweight'] = total_grams
            out.append(aggregate)
        return out

    def build_item(self, product, sender, recipient, weight_grams, amount=0, currency='EUR',
                   shipment_nature='SALE_GOODS', customer_reference=None, contents=None):
        """Build an 'item' in the language of the WaPoInt API. Represents one shipment."""
        weight_grams = int(weight_grams)
        if weight_grams > 2000:
            raise ValueError('Maximum item gross weight is 2000g')
        if len(currency) != 3:
            raise ValueError('Currency must be expressed as 3-digit ISO-4217 code')
        if contents and len(contents) > 5:
            raise ValueError('Custom Contents must not contain more than 5 lines')
        ret = {
            'product': str(product),
            'serviceLevel': 'STANDARD',
            'shipmentAmount': int(amount),
            'shipmentCurrency': currency,
            'shipmentGrossWeight': weight_grams,
            'shipmentNaturetype': shipment_nature,
            }
        # merge in the sender and recipient fields
        ret.update(sender.as_sender())
        ret.update(recipient.as_recipient())
        if contents:
            ret['contents'] = contents
        if customer_reference:
            customer_reference = str(customer_reference)
            if len(customer_reference) > 20:
                raise ValueError('Maximum length of customer reference is 20 chars')
            ret['custRef'] = customer_reference
        return ret

    def build_order(self, items, contact_name, order_status='FINALIZE'):
        """Build an 'order' in the language of the WaPoInt API.  Consists of multiple shipments."""
        ret = {
            'customerEkp': self.ekp,
            'orderStatus': order_status,
            'paperwork': {
                'contactName': contact_name,
                'awbCopyCount': 1,
                },
            'items': items
            }
        return ret

    def api_create_order(self, items, contact_name, order_status='FINALIZE'):
        """Issue an API request to create an order consisting of items."""
        order = self.build_order(items, contact_name=contact_name, order_status=order_status)
        _logger.info("Order Request: %s", order)
        r = self.request('POST', 'orders', json = order)
        if r.ok:
        # TODO: figure out the AWB and the (item, barcode, voucherId, ...) for the items
            json_resp = r.json()
            #print(json.dumps(json_resp, indent=4))
            _logger.info("Order Response: %s", json_resp)
            # TODO: download the PDF for each AWB
            return json_resp
        else:
            raise ValueError('%s: %s' % (r.status_code, r.text))

    def api_get_item_label(self, item_id, accept='application/pdf'):
        """Download the label for a given item.  Returns PDF as 'bytes'"""
        r = self.request('GET', 'items/%s/label' % item_id, headers={'Accept': accept})
        r.raise_for_status()
        return r.content

    def api_get_item_labels(self, awb, accept='application/pdf'):
        """Download the labels for all items in a given AWB. Returns PDF as 'bytes'"""
        r = self.request('GET', 'shipments/%s/itemlabels' % awb, headers={'Accept': accept})
        r.raise_for_status()
        return r.content



#!/usr/bin/env python3


# SPDX-License-Identifier: LGPL-3.0-or-later
# SPDX-FileCopyrightText: © 2025 Georg Sauthoff <mail@gms.tf>


from decimal import Decimal
import magic
import pdfminer.high_level as pmhl
import pytest
import requests
import tomllib
import zipfile


import inema.rest as ir


@pytest.fixture(scope='session')
def im():
    with open('test.toml', 'rb') as f:
        d = tomllib.load(f)
    a = d['account']
    im = ir.Session(a['client_id'], a['client_secret'], a['user'], a['password'])
    return im


def test_health():
    d = ir.check_health()
    assert d['name'] == 'pp-post-internetmarke'
    assert d['version'][0] == 'v'
    assert 'rev' in d
    assert 'env' in d


def test_get_formats(im):
    fs = im.get_formats()
    assert len(fs) > 100
    xs = [ x for x in fs if x['id'] == 5 ]
    assert len(xs) == 1
    x = xs[0]
    assert x['name'] == 'Brief C6 114 x 162'
    assert x['pageLayout']['size']['x'] == 162


def test_user_profile(im):
    d = im.profile()
    assert 'company' in d
    assert 'firstname' in d
    assert 'lastname' in d


def test_charge(im):
    old_balance = im.balance
    d = im.charge(3000)
    assert d['walletBalance'] == im.balance
    assert old_balance + 3000 == im.balance


def test_buy_pdf(im, tmp_path):
    old_balance = im.balance

    oid = im.create_order()
    p = ir.mk_pdf_pos(1022,
            sender=ir.mk_addr('Erika Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernd Müller', 'Nußhäherstraße 10', '12345', 'München')   )
    t = ir.compute_total(p, im)

    body = ir.mk_pdf_req(oid, 5, p, t)
    fn   = tmp_path / 'postage.pdf'
    d    = im.checkout_pdf(body, fn)

    assert d['shoppingCart']['shopOrderId'] == oid
    assert im.balance != old_balance

    s = pmhl.extract_text(fn)

    xs = s.strip().splitlines()
    assert xs[0] == 'Erika Mustermann, Heidestr. 17, 51147 Köln'
    assert xs[4].strip() == 'REINSCHREIBEN  EINWURF'
    assert xs[-3] == 'Bernd Müller'
    assert xs[-2] == 'Nußhäherstraße 10'
    assert xs[-1] == '12345 München'

    x = [ x for x in xs if x.startswith('IM ') ][0].split()[-1].replace(',', '.')
    d = int(Decimal(x) * Decimal('100'))
    assert d == old_balance - im.balance



def test_cancel(im):
    old_balance = im.balance

    oid = im.create_order()
    p = ir.mk_pdf_pos(1, layout='FRANKING_ZONE')
    t = ir.compute_total(p, im)

    body = ir.mk_pdf_req(oid, 5, [p], t)
    d    = im.checkout_pdf(body)

    delta = old_balance - im.balance
    vid = d['shoppingCart']['voucherList'][0]['voucherId']

    assert im.balance != old_balance

    assert len(d['link']) > 0

    b = im.ses.get(d['link']).content

    d = im.cancel(ir.mk_cancel_req(order_id=oid))
    rid = d['shopRetoureId']
    tid = d['retoureTransactionId']

    d = im.list_cancel(retoure_id=rid)
    assert len(d) == 1
    d = d[0]

    assert d['totalCount'] == 1
    assert d['countStillOpen'] == 0
    assert d['retourePrice'] == delta
    assert len(d['refundedVouchers']) == 1
    assert d['refundedVouchers'][0]['voucherId'] == vid



def test_buy_pdf_multi(im, tmp_path):
    old_balance = im.balance

    oid = im.create_order()
    ps = [
        ir.mk_pdf_pos(11, 1, 1, 1,
            sender=ir.mk_addr('Max Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernd Brot', 'Nußhäherstraße 10', '12345', 'München') ),
        ir.mk_pdf_pos(1, 2, 1, 1,
            sender=ir.mk_addr('Erika Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernadine Müller', 'Nußhäherstraße 10', '12345', 'München') ),
        ir.mk_pdf_pos(31, 2, 2, 1),
        ir.mk_pdf_pos(290, 1, 2, 1,
            sender=ir.mk_addr('Joe Doe', 'Heideweg 23', '12345', 'Köln'),
            receiver=ir.mk_addr('Wolfgang Göthe', 'Nußhäherstraße 10', '12345', 'München') ),
    ]
    t = ir.compute_total(ps, im)

    body = ir.mk_pdf_req(oid, 1, ps, t)
    fn   = tmp_path / 'postage.pdf'
    d    = im.checkout_pdf(body, fn)

    assert d['shoppingCart']['shopOrderId'] == oid
    # http has referer and Internetmarke has Ballance [sic] ...
    assert im.balance == d['walletBallance']
    assert im.balance != old_balance
    assert im.balance + t == old_balance
    assert len(d['shoppingCart']['voucherList']) == 4
    old_link = d['link']
    old_vl = d['shoppingCart']['voucherList']

    d = im.list_order(oid)

    assert d['shoppingCart']['shopOrderId'] == oid
    assert len(d['shoppingCart']['voucherList']) == 4
    assert d['shoppingCart']['voucherList'] == old_vl
    assert len(d['link']) > 0
    # usually new link, but same postage, wouldn't need to be different, though ...
    # assert d['link] != old_link


def test_buy_png(im, tmp_path):
    old_balance = im.balance

    oid = im.create_order()
    p = ir.mk_png_pos(1022,
            sender=ir.mk_addr('Erika Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernd Müller', 'Nußhäherstraße 10', '12345', 'München')   )
    t = ir.calc_total(p)

    body = ir.mk_png_req(oid, p, t)
    fn   = tmp_path / 'postage-png.zip'
    d    = im.checkout_png(body, fn)

    assert d['shoppingCart']['shopOrderId'] == oid
    assert im.balance != old_balance
    assert im.balance + t == old_balance

    m = magic.detect_from_filename(fn)
    assert m.mime_type == 'application/zip'

    z = zipfile.ZipFile(fn)
    xs = [ x.filename for x in z.infolist() ]
    assert xs == [ '0.png' ]


def test_buy_png_multi(im, tmp_path):
    old_balance = im.balance

    oid = im.create_order()
    ps = [
        ir.mk_png_pos(1022,
            sender=ir.mk_addr('Erika Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernd Müller', 'Nußhäherstraße 10', '12345', 'München')),
        ir.mk_png_pos(1002,
            sender=ir.mk_addr('Peter Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Petra Müller', 'Nußhäherstraße 10', '12345', 'München')),
    ]
    h = {}
    t = ir.compute_total(ps, im, h)

    assert len(h) == 2
    assert 1022 in h
    assert 1002 in h

    body = ir.mk_png_req(oid, ps, t)
    fn   = tmp_path / 'postage-png.zip'
    d    = im.checkout_png(body, fn)

    assert d['shoppingCart']['shopOrderId'] == oid
    assert im.balance != old_balance
    assert im.balance + t == old_balance

    m = magic.detect_from_filename(fn)
    assert m.mime_type == 'application/zip'

    z = zipfile.ZipFile(fn)
    xs = [ x.filename for x in z.infolist() ]
    assert xs == [ '0.png', '1.png' ]


def test_preview_png(im, tmp_path):
    fn   = tmp_path / 'postage.png'

    d = im.preview_png(ir.mk_png_preview_req(10001), fn)
    m = magic.detect_from_filename(fn)
    assert m.mime_type == 'image/png'


def test_remote_price_error(im):
    old_balance = im.balance
    oid = im.create_order()
    p = ir.mk_pdf_pos(1,
            sender=ir.mk_addr('Erika Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernd Müller', 'Nußhäherstraße 10', '12345', 'München')   )

    total = 1
    body = ir.mk_pdf_req(oid, 1, p, total)

    with pytest.raises(requests.exceptions.HTTPError) as e:
        im.checkout_pdf(body)

    d = e.value.response.json()
    assert d['statusCode'] == '400'
    assert d['title'] == 'invalidTotalAmount'
    assert d['description'].startswith('The total amount of the order is invalid!')
    xs = d['description'].split()
    assert xs[-4] == 'PPL-total:'
    assert xs[-3] == '95,'
    assert xs[-2] == 'Cart:'
    assert xs[-1] == '1'
    assert im.balance == old_balance



def test_remote_format_error(im):
    old_balance = im.balance
    oid = im.create_order()
    p = ir.mk_pdf_pos(1,
            sender=ir.mk_addr('Erika Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernd Müller', 'Nußhäherstraße 10', '12345', 'München')   )
    p['position']['labelX'] = 0

    # NB: as of 2025-09, web service doesn't check for colum/row overflow ...

    t = ir.compute_total(p, im)
    body = ir.mk_pdf_req(oid, 1, p, t)

    with pytest.raises(requests.exceptions.HTTPError) as e:
        im.checkout_pdf(body)

    d = e.value.response.json()
    assert d['statusCode'] == '400'
    assert d['title'] == 'Bad Request'
    assert im.balance == old_balance






#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2016 Georg Sauthoff <mail@georg.so>
# SPDX-License-Identifier: LGPL-3.0-or-later

import argparse
import configparser
import csv
import datetime
import json
import logging
import os
import re
import requests
import subprocess
import sys
import zeep
import zeep.helpers

# to use a developer version of inema if available
if __name__ == '__main__':
  import inspect
  s = inspect.getsourcefile(lambda:0)
  if s and s != '<stdin>':
    d = os.path.dirname(os.path.abspath(s)) + '/python-inema'
    if os.path.exists(d):
      sys.path.insert(0, d)
  from inema.data import products, formats
  from inema.soapfree import Internetmarke
else:
  from .data import products, formats
  from .soapfree import Internetmarke

class Fake_IM:

  def checkoutPDF(self, format_id):
    pass

  def build_addr(self, street, number, code, city, country):
    pass

  def build_pers_addr(self, first, name, address):
    pass

  def build_comp_addr(self, first, name, address):
    pass

  def build_position(self, product, sender, receiver, layout = "AddressZone", pdf = False, x=1, y=1, page=1):
    pass

  def add_position(self, position):
    pass

  def retrievePreviewPDF(self, prod_code, page_format, layout = "AddressZone"):
    pass

  def compute_total(self):
    return 0

try:
  import colorlog
except ImportError:
  pass

log_format    = '%(asctime)s - %(levelname)-8s - %(message)s [%(name)s]'
log_date_format = '%Y-%m-%d %H:%M:%S'

def setup_logging():
  root = logging.getLogger()
  root.setLevel(logging.WARN)
  logging.getLogger(__name__).setLevel(logging.INFO)

  if 'colorlog' in sys.modules and os.isatty(2):
    cformat   = '%(log_color)s' + log_format
    cf = colorlog.ColoredFormatter(cformat, log_date_format,
      log_colors = { 'DEBUG': 'reset', 'INFO': 'reset',
        'WARNING' : 'bold_yellow' , 'ERROR': 'bold_red',
        'CRITICAL': 'bold_red'})

  else:
    cf = logging.Formatter(log_format, log_date_format)

  ch = logging.StreamHandler()
  ch.setFormatter(cf)
  root.addHandler(ch)

log = logging.getLogger(__name__)

def setup_file_logging(filename):
  root = logging.getLogger()
  root.setLevel(logging.DEBUG)
  logging.getLogger(__name__).setLevel(logging.NOTSET)

  fh = logging.FileHandler(filename)
  fh.setLevel(logging.DEBUG)
  f = logging.Formatter(log_format, log_date_format)
  fh.setFormatter(f)
  root.addHandler(fh)

  class Filter:
    def filter(self, r):
      return r.name == __name__ or r.levelno >= logging.WARNING

  ch = root.handlers[0]
  ch.setLevel(logging.INFO)
  ch.addFilter(Filter())



def mk_arg_parser():
  p = argparse.ArgumentParser(
      formatter_class=argparse.RawDescriptionHelpFormatter,
      description='buy postage online',
      epilog='''The program interfaces with the Deutsche Post
postage service for letters and small packages.

Account details are read from a config file, by default this is
~/.config/frank.conf. An example config:

    [api]
    id = your-partner-id
    key = your-api-key
    key_phase = 1

    [account]
    user = portokasse-user
    password = portokasse-pw

    [a.default]
    first =
    name = Firma ACME
    street = Lindenallee
    number = 3
    zip = 12345
    city = Bielefeld

Examples:

List all formats that have a height of 297 mm:

    $ frank --list-formats x297

List all products that are called 'sendung' or so:

   $ frank --list-products sendung

Preview a Büchersendung stamp (creates postage_YYYY-MM-DD.pdf):

   $ frank --preview --product 78 --format 1

Frank and buy 2 stamps (creates 2 page document postage_YYYY-MM-DD.pdf):

   $ frank ---format 26  --product 79 'Joe User;Street 1;12345 City' \\
       'Jane User;Fakestreet 2;67890 Fakestadt'

It's also fine to delimit the recipient lines with newline characters.

2016, Georg Sauthoff <mail@georg.so>, GPLv3+

'''
      )
  p.add_argument('recipients', metavar='RECIPIENT', nargs='*',
      help = 'recipients')
  p.add_argument('--config', action='append',
      metavar='FILENAME', help='user specific config file')
  p.add_argument('--csv', metavar='FILENAME',
      help='read recipient data from CSV file (1st row is header)')
  p.add_argument('--debug', help='store debug message into log file')
  p.add_argument('--dry', action='store_true', help='dry run')
  p.add_argument('--format' ,'-f', default='1',
      help='format id for the resulting pdf')
  p.add_argument('--global-conf', default = '/usr/share/frank/frank.conf',
      metavar='FILENAME', help='global config file')
  p.add_argument('--json', action='store_true',
      help='print tablses as json')
  p.add_argument('--list-formats', nargs='?', default=None, const='.',
      metavar='REGEX', help='list available formats')
  p.add_argument('--list-products', nargs='?', default=None, const='.',
      metavar='REGEX', help='list available products')
  p.add_argument('--manifest', action='store_true',
      help='write manifest pdf')
  p.add_argument('--output', '-o', default='.', metavar='DIRECTORY',
      help='output directory where postage files are created')
  p.add_argument('--preview', action='store_true',
      help='only retrieve preview documents')
  p.add_argument('--product' ,'-p', action='append',
      help='product id(s) to use for the recipient(s)')
  p.add_argument('--sender', action='append', help='sender(s)')
  p.add_argument('--suffix', default='', help='postage basename suffix')
  p.add_argument('--sys-conf', default = '/etc/frank.conf',
      metavar='FILENAME', help='machine specific config file')
  p.add_argument('--update', action='store_true',
      help='update internal format list via webservice')
  p.add_argument('--print', action='store_true', default=False,
      help='Print the retrieved PDF with the lpr command')
  p.add_argument('--version', '-v', action='store_true',
      help='Display inema package version')
  return p

def parse_args(*xs):
  arg_parser = mk_arg_parser()
  args = arg_parser.parse_args(*xs)

  if args.version:
      import importlib.metadata
      try:
          print(importlib.metadata.version('inema'))
      except importlib.metadata.PackageNotFoundError:
          print('not-installed')
      sys.exit(0)

  if args.debug:
    setup_file_logging(args.debug)
  if not args.config:
    args.config = [ '~/.config/frank.conf' ]
  if not args.format:
    args.format = ['26']
  if not args.sender:
    args.sender = ['$default']
  return args

def read_config(filenames):
  c = configparser.ConfigParser()
  c.read(filenames)
  return c

def list_products(expr):
  l = []
  e = re.compile(expr, flags=re.IGNORECASE)
  for k,v in products.items():
    if not e.search(v['name']):
      continue
    h = v
    h['id'] = int(k)
    if not h['max_weight']:
      h['max_weight'] = ''
    s = h['cost_price']
    t = s.split('.')
    if t.__len__() == 1:
      h['cost_price'] = s + ' ' * 3
    else:
      h['cost_price'] = s + ' ' * (2 - t[1].__len__())
    l.append(h)
  l.sort(key=lambda h : h['id'])
  fs = '{:>6} {:<70} {:>6} {:>6} {:>5}'
  print(fs.format('id', 'name', 'EUR', 'g', 'intl'))
  print('-'*(6+70+6+6+5 +4))
  for h in l:
    print(fs.format(h['id'], h['name'], h['cost_price'], h['max_weight'],
      h['international']))

def list_formats(expr):
  e = re.compile(expr, flags=re.IGNORECASE)
  fs = '{:>6} {:<44} {:>3}*{:>3} {:>5} {:<12} {:>3} {:>3}'
  print(fs.format('id', 'name', 'w', 'h', '#ls', 'type', 'adr', 'img'))
  print('-'*(6+44+3+3+5+12+3+3 +7))
  for f in formats:
    if e.search(f['name']) or e.search(f['pageType']) \
        or e.search('{}x{}'.format(f['pageLayout']['size']['x'],
                                   f['pageLayout']['size']['y'])):
      print(fs.format(f['id'], f['name'], int(f['pageLayout']['size']['x']),
          int(f['pageLayout']['size']['y']),
           int(f['pageLayout']['labelCount']['labelX'])
          *int(f['pageLayout']['labelCount']['labelY']), f['pageType'],
          f['isAddressPossible'], f['isImagePossible']) )

def parse_address(s, conf):
  if s.startswith('$'):
    h = conf['a.'+s[1:]]
    return ( h.get('first', ''), h['name'], h['street'], h['number'],
        h['zip'], h['city'], h.get('country', 'DEU') )
  delimiter = None
  for d in ['\n', ';']:
    if d in s:
      delimiter = d
  if not delimiter:
    raise ValueError('recipient string has no known delimiters')
  first = ''; name = ''; street = ''; number = ''; zipcode = ''; city = ''
  country = 'DEU'
  l = s.split(delimiter)
  if l.__len__() > 0:
    xs = l[0].split(' ')
    if xs.__len__() == 1:
      first = ''
      name = xs[0]
    else:
      first = ' '.join(xs[0:-1])
      name = xs[-1]
  if l.__len__() > 1:
    xs = l[1].split(' ')
    if xs.__len__() == 1:
      street = xs[0]
      number = ''
    else:
      street = ' '.join(xs[0:-1])
      number = xs[-1]
  if l.__len__() > 2:
    xs = l[2].split(' ')
    if xs.__len__() == 1:
      zipcode = ''
      city = xs[0]
    else:
      zipcode = xs[0]
      city = ' '.join(xs[1:])
  if l.__len__() > 3:
    c = l[3].strip()
    if c:
      country = c
  return (first, name, street, number, zipcode, city, country)

def parse_csv(filename):
  xs = []
  ps = []
  with open(filename, 'r') as f:
    rs = csv.reader(f)
    next(rs)
    for r in rs:
      xs.append(r[0:7] + ['']*(7-r.__len__()))
      if r.__len__() > 7:
        ps.append(r[7])
  return (xs, ps)


def parse_addresses(args, conf):
  recipients = []
  for r in args.recipients:
    recipients.append(parse_address(r, conf))
  args.recipients = recipients
  sender = []
  for r in args.sender:
    sender.append(parse_address(r, conf))
  args.sender = sender
  if args.csv:
    t = parse_csv(args.csv)
    args.recipients = args.recipients + t[0]
    if not args.product:
      args.product = []
    args.product = args.product + t[1]

def apply_config(args, conf):
  if not args.manifest and conf.has_section('general'):
    args.manifest = conf['general'].get('manifest', False)

def mk_address(im, x, conf):
  a = im.build_addr(x[-5], x[-4], x[-3], x[-2], x[-1])
  if x[0]:
    r = im.build_pers_addr(x[0], x[1], a)
  else:
    r = im.build_comp_addr(x[1], a)
  return r;

def buy(im, sender, recipient, product, i, pi, args, conf):
  src = mk_address(im, sender, conf)
  dst = mk_address(im, recipient, conf)
  page   = int(i / (pi[0] * pi[1])) + 1
  column = int(i % pi[0]) + 1
  row    = int(int(i / pi[0]) % pi[1]) + 1
  p = im.build_position(product, src, dst, pdf=True,
      page=page, x=column, y=row)
  im.add_position(p)


def mk_filename(args, base='postage'):
  filename = '{}/{}_{}{}.pdf'.format(args.output, base,
      datetime.datetime.now().strftime('%Y-%m-%d'), args.suffix)
  return filename

def store_files(res, args):
  if hasattr(res, 'manifest_pdf_bin') and args.manifest:
    with open(mk_filename(base='manifest'), 'wb') as f:
      f.write(res.manifest_pdf_bin)
  i = 1
  pdf_bin = None
  if hasattr(res, 'pdf_bin'):
    pdf_bin = res.pdf_bin
  elif hasattr(res.shoppingCart.voucherList.voucher[0], 'pdf_bin'):
    pdf_bin = res.shoppingCart.voucherList.voucher[0].pdf_bin
  if pdf_bin:
    filename = mk_filename(args)
    log.info('Writing: {}'.format(filename))
    with open(filename, 'wb') as f:
      f.write(pdf_bin)
    if args.print:
      log.info('Printing: {}'.format(filename))
      subprocess.check_call(['lpr', filename])

def get_format(ident):
  for f in formats:
    if f['id'] == int(ident):
      return f
  raise ValueError("Couldn't find format id: {}".format(ident))

def get_page_info(f):
  return ( int(f['pageLayout']['labelCount']['labelX']),
           int(f['pageLayout']['labelCount']['labelY']) )

def do_list_products(args):
  if args.list_products:
    list_products(args.list_products)
    return True

def do_list_formats(args):
  if args.list_formats:
    if args.json:
      print(json.dumps(formats, indent=2))
    else:
      list_formats(args.list_formats)
    return True

def do_update_list_formats(im, args):
  if args.list_formats and args.update:
    formats = sorted(zeep.helpers.serialize_object(
      im.retrievePageFormats()), key=lambda x:x['id'])
    return do_list_formats(args)

def do_create_preview(im, args):
  if args.preview:
    link = im.retrievePreviewPDF(args.product[0], args.format)
    pdf = requests.get(link, stream=True)
    with open(mk_filename(args), 'wb') as f:
      f.write(pdf.content)
    return True

def run(args, conf):
  if do_list_products(args):
    return 0
  if not args.update and do_list_formats(args):
    return 0
  ps = args.product
  ss = args.sender
  if args.dry:
    im = Fake_IM()
  else:
    im = Internetmarke(conf['api']['id'], conf['api']['key'],
        conf['api'].get('key_phase', '1'))
    im.authenticate(conf['account']['user'], conf['account']['password'])
  if do_create_preview(im, args):
    return 0
  if do_update_list_formats(im, args):
    return 0
  page = get_page_info(get_format(args.format))
  i = 0
  for r in args.recipients:
    buy(im, ss[0], r, ps[0], i, page, args, conf)
    if ps.__len__() > 1:
      ps = ps[1:]
    if ss.__len__() > 1:
      ss = ss[1:]
    i = i + 1
  log.warning('Buying postage for {} €'.format(im.compute_total()/100))
  res = im.checkoutPDF(args.format)
  store_files(res, args)
  return 0

def imain(args):
  conf = read_config([args.global_conf, args.sys_conf]
    + [os.path.expanduser(x) for x in args.config] )
  parse_addresses(args, conf)
  apply_config(args, conf)
  try:
    return run(args, conf)
  except zeep.exceptions.Fault as e:
    d = str(e.detail)
    try:
      d = zeep.wsdl.utils.etree_to_string(e.detail).decode()
      ids = e.detail.xpath('//*[name()="id"]')
      ms = e.detail.xpath('//*[name()="message"]')
      d = " - ".join(", ".join(x.text for x in xs) for xs in (ids, ms))
    except TypeError:
      pass
    log.error('{} ({})'.format(e.message, d))
    return 1

def main():
  setup_logging()
  args = parse_args()
  log.debug('Starting frank.py')
  return imain(args)

if __name__ == '__main__':
  sys.exit(main())




#!/usr/bin/env python3


# SPDX-License-Identifier: LGPL-3.0-or-later
# SPDX-FileCopyrightText: © 2025 Georg Sauthoff <mail@gms.tf>


import pytest

import json
import jsonschema
import referencing

import inema.rest as ir


# json schema files can be extracted out of the Internetmarke OpenAPI specification,
# e.g. with https://github.com/instrumenta/openapi2jsonschema (o2js)
#     
# as of 2025, o2js output needs case fixing of references:
#
#     sed '/"\$ref":/s/.*/\L&/' path/to/*.json -i
#
# cf. https://github.com/instrumenta/openapi2jsonschema/issues/63
schema_base = 'schema/rest'


def json_lookup(s):
    assert s.endswith('.json')
    assert '/' not in s
    # does not auto detect schema ...
    # return referencing.Resource.from_contents(json.load(open(f'{schema_base}/{s}')))
    schema = json.load(open(f'{schema_base}/{s}'))
    jsonschema.Draft202012Validator.check_schema(schema)
    return referencing.jsonschema.DRAFT202012.create_resource(schema)


@pytest.fixture(scope='session')
def reg():
    r = referencing.Registry(retrieve=json_lookup)
    return r


def test_addr(reg):
    schema = json_lookup('address.json').contents
    v = jsonschema.Draft202012Validator(schema, registry=reg)
    d = ir.mk_addr('Erika Mustermann', 'Heidestr. 17', '51147', 'Köln')
    v.validate(d)

def test_pdf_pos(reg):
    schema = json_lookup('appshoppingcartpdfposition.json').contents
    v = jsonschema.Draft202012Validator(schema, registry=reg)
    d = ir.mk_pdf_pos(1022,
            sender=ir.mk_addr('Erika Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernd Müller', 'Nußhäherstraße 10', '12345', 'München')   )
    v.validate(d)

def test_png_pos(reg):
    schema = json_lookup('appshoppingcartposition.json').contents
    v = jsonschema.Draft202012Validator(schema, registry=reg)
    d = ir.mk_png_pos(1022,
            sender=ir.mk_addr('Erika Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernd Müller', 'Nußhäherstraße 10', '12345', 'München')   )
    v.validate(d)

def test_pdf_req(reg):
    schema = json_lookup('appshoppingcartpdfrequest.json').contents
    v = jsonschema.Draft202012Validator(schema, registry=reg)
    p = ir.mk_pdf_pos(1022,
            sender=ir.mk_addr('Erika Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernd Müller', 'Nußhäherstraße 10', '12345', 'München')   )
    oid = 42
    t = 23
    d = ir.mk_pdf_req(oid, 5, p, t)
    v.validate(d)

def test_pdf_req_multi(reg):
    schema = json_lookup('appshoppingcartpdfrequest.json').contents
    v = jsonschema.Draft202012Validator(schema, registry=reg)
    ps = [
        ir.mk_pdf_pos(11, 1, 1, 1,
            sender=ir.mk_addr('Max Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernd Brot', 'Nußhäherstraße 10', '12345', 'München') ),
        ir.mk_pdf_pos(1, 2, 1, 1,
            sender=ir.mk_addr('Erika Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernadine Müller', 'Nußhäherstraße 10', '12345', 'München') ),
        ir.mk_pdf_pos(31, 2, 2, 1),
        ir.mk_pdf_pos(290, 1, 2, 1,
            sender=ir.mk_addr('Joe Doe', 'Heideweg 23', '12345', 'Köln'),
            receiver=ir.mk_addr('Wolfgang Göthe', 'Nußhäherstraße 10', '12345', 'München') ),
    ]
    oid = 42
    t = 23
    d = ir.mk_pdf_req(oid, 1, ps, t)
    v.validate(d)

def test_png_req(reg):
    schema = json_lookup('appshoppingcartpngrequest.json').contents
    v = jsonschema.Draft202012Validator(schema, registry=reg)
    p = ir.mk_png_pos(1022,
            sender=ir.mk_addr('Erika Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernd Müller', 'Nußhäherstraße 10', '12345', 'München')   )
    oid = 42
    t = 23
    d = ir.mk_png_req(oid, p, t)
    v.validate(d)

def test_png_req_multi(reg):
    schema = json_lookup('appshoppingcartpngrequest.json').contents
    v = jsonschema.Draft202012Validator(schema, registry=reg)
    ps = [
        ir.mk_png_pos(1022,
            sender=ir.mk_addr('Erika Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernd Müller', 'Nußhäherstraße 10', '12345', 'München')),
        ir.mk_png_pos(1002,
            sender=ir.mk_addr('Peter Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Petra Müller', 'Nußhäherstraße 10', '12345', 'München')),
    ]
    oid = 42
    t = 23
    d = ir.mk_png_req(oid, ps, t)
    v.validate(d)

def test_cancel_req(reg):
    schema = json_lookup('retourevouchersrequest.json').contents
    v = jsonschema.Draft202012Validator(schema, registry=reg)
    oid = 42
    d = ir.mk_cancel_req(order_id=oid)
    v.validate(d)

def test_pdf_preview_req(reg):
    schema = json_lookup('appshoppingcartpreviewpdfrequest.json').contents
    v = jsonschema.Draft202012Validator(schema, registry=reg)
    d = ir.mk_pdf_preview_req(5, 31)
    v.validate(d)

def test_png_preview_req(reg):
    schema = json_lookup('appshoppingcartpreviewpngrequest.json').contents
    v = jsonschema.Draft202012Validator(schema, registry=reg)
    d = ir.mk_png_preview_req(290)
    v.validate(d)






import json

import importlib.resources




def load_products():
    """Load bundled products dictionary.

    Read the most recent inema/data/products-YYYY-MM-DD.json where YYYY-MM-DD is
    not more recent than today. Fall back to inema/data/products.json.
    This allows to ship multiple products.json files for announced future price/product
    changes.

    Returns
    -------
    dict
        Mapping of product code string to product information dictionary that provides cost_price and other fields.

    Warning
    -------
    The `cost_price` field is in euro while all price amounts in the Internetmarke REST API are in euro cent.
    """
    ps = [ e for e in (x.name for x in (importlib.resources.files(__package__) / 'data').iterdir())
            if e.startswith('products-') and e.endswith('.json')
            and e <= 'products-{}.json'.format(date.today().isoformat()) ]
    ps.sort()
    ps.insert(0, 'products.json')
    x = importlib.resources.files('inema') / 'data' / ps[-1]
    t = x.read_text(encoding='utf-8')
    products = json.loads(t)
    return products

products = load_products()
"""Default product dictionary.

Initialized with the result of :func:`load_products`.

Example
-------
>>> inema.data.products['1']
{'cost_price': '0.95',
 'international': False,
 'max_weight': '20',
 'name': 'Standardbrief'}

:meta hide-value:
"""


def load_formats():
    """Load bundled format list.

    Returns
    -------
    list
        List of format dictionaries.

    """
    x = importlib.resources.files(__package__) / 'data' / 'formats.json'
    t = x.read_text(encoding='utf-8')
    formats = json.loads(t)
    return formats

formats = load_formats()
"""Default format list.

Initialized with the result of :func:`load_formats`.

Example
-------
>>> inema.data.formats[0]
Out[2]:
{'id': 1,
 'isAddressPossible': True,
 'isImagePossible': False,
 'name': 'DIN A4 Normalpapier',
 'description': None,
 'pageType': 'REGULARPAGE',
 'pageLayout': {'size': {'x': 210, 'y': 297},
  'orientation': 'PORTRAIT',
  'labelSpacing': {'x': 0, 'y': 0},
  'labelCount': {'labelX': 2, 'labelY': 5},
  'margin': {'top': 31, 'bottom': 31, 'left': 15, 'right': 15}}}

:meta hide-value:
"""




#!/usr/bin/env python3


# SPDX-License-Identifier: LGPL-3.0-or-later
# SPDX-FileCopyrightText: © 2025 Georg Sauthoff <mail@gms.tf>


# This test suite validates REST requests constructed by inema.rest.Session
# against the Internetmarke Openapi specification published by Deutsche Post.
#
# It doesn't require internet access, just a copy of that spec file.
#
# See also the Downloads Section in the Internetmarke documentation:
# https://developer.dhl.com/api-reference/deutsche-post-internetmarke-post-parcel-germany#downloads-section


import pytest

import openapi_core
import openapi_core.contrib.requests
import requests
import responses
import yaml


import inema.rest as ir



inema_openapi_filename = 'pp-post-internetmarke-latest.yaml'
inema_url = 'http://localhost:6666/v1/'

# otherwise, openapi runs into discriminator property based referencing errors,
# during validation ...
#
# NB: an allOf-adjacent discriminator object isn't covered by the openapi spec
def fix_bogus_discrimi(y):
    for k, v in y['components']['schemas'].items():
        if 'allOf' in v and 'discriminator' in v:
            v.pop('discriminator')


@pytest.fixture(scope='session')
def oa():
    with open(inema_openapi_filename) as f:
        y = yaml.safe_load(f)
        fix_bogus_discrimi(y)
        y['servers'][0]['url'] = inema_url[:-1]
        oa = openapi_core.OpenAPI.from_dict(y)
        return oa


class Early_Exit(Exception):
    pass

class Fake_Session:
    def __init__(self, headers):
        self.headers = headers
        self.req = None

    def post(self, *xs, **kw):
        kw['headers'] = self.headers
        self.req = requests.Request('POST', *xs, **kw)
        return self

    def get(self, *xs, **kw):
        kw['headers'] = self.headers
        self.req = requests.Request('GET', *xs, **kw)
        return self

    def put(self, *xs, **kw):
        kw['headers'] = self.headers
        self.req = requests.Request('PUT', *xs, **kw)
        return self

    def raise_for_status(self):
        raise Early_Exit()

    def clear(self):
        self.req = None


@pytest.fixture(scope='session')
def im():
    with responses.RequestsMock() as r:
        base = inema_url
        r.add(responses.POST,
              base + 'user',
              body='{"access_token": "0815","walletBalance":230}',
              status=200,
              content_type='application/json',
              )
        s = ir.Session('client', 'geheim', 'juser', 'einsfueralles', api_base=base)
        s.ses = Fake_Session(s.ses.headers)
        return s


def test_user_profile(im, oa):
    im.ses.clear()
    try:
        im.profile()
    except Early_Exit:
        pass
    r = im.ses.req
    ro = openapi_core.contrib.requests.RequestsOpenAPIRequest(r)
    oa.validate_request(ro)


def test_get_format(im, oa):
    im.ses.clear()
    try:
        im.get_formats()
    except Early_Exit:
        pass
    r = im.ses.req
    ro = openapi_core.contrib.requests.RequestsOpenAPIRequest(r)
    oa.validate_request(ro)


def test_charge(im, oa):
    im.ses.clear()
    try:
        im.charge(3000)
    except Early_Exit:
        pass
    r = im.ses.req
    ro = openapi_core.contrib.requests.RequestsOpenAPIRequest(r)
    oa.validate_request(ro)


def test_create_order(im, oa):
    im.ses.clear()
    try:
        im.create_order()
    except Early_Exit:
        pass
    r = im.ses.req
    ro = openapi_core.contrib.requests.RequestsOpenAPIRequest(r)
    oa.validate_request(ro)


def test_buy_pdf(im, oa):
    im.ses.clear()
    oid = 123
    p = ir.mk_pdf_pos(1022,
            sender=ir.mk_addr('Erika Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernd Müller', 'Nußhäherstraße 10', '12345', 'München')   )
    t = 100
    body = ir.mk_pdf_req(oid, 5, p, t)
    try:
        im.checkout_pdf(body)
    except Early_Exit:
        pass
    r = im.ses.req
    ro = openapi_core.contrib.requests.RequestsOpenAPIRequest(r)
    oa.validate_request(ro)


def test_cancel(im, oa):
    im.ses.clear()
    oid = 123
    d = ir.mk_cancel_req(order_id=oid)
    try:
        im.cancel(d)
    except Early_Exit:
        pass
    r = im.ses.req
    ro = openapi_core.contrib.requests.RequestsOpenAPIRequest(r)
    oa.validate_request(ro)


def test_list_cancel(im, oa):
    im.ses.clear()
    rid = 123
    try:
        im.list_cancel(retoure_id=rid)
    except Early_Exit:
        pass
    r = im.ses.req
    ro = openapi_core.contrib.requests.RequestsOpenAPIRequest(r)
    oa.validate_request(ro)


def test_buy_pdf_multi(im, oa):
    im.ses.clear()
    oid = 123
    t = 100
    ps = [
        ir.mk_pdf_pos(11, 1, 1, 1,
            sender=ir.mk_addr('Max Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernd Brot', 'Nußhäherstraße 10', '12345', 'München') ),
        ir.mk_pdf_pos(1, 2, 1, 1,
            sender=ir.mk_addr('Erika Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernadine Müller', 'Nußhäherstraße 10', '12345', 'München') ),
        ir.mk_pdf_pos(31, 2, 2, 1),
        ir.mk_pdf_pos(290, 1, 2, 1,
            sender=ir.mk_addr('Joe Doe', 'Heideweg 23', '12345', 'Köln'),
            receiver=ir.mk_addr('Wolfgang Göthe', 'Nußhäherstraße 10', '12345', 'München') ),
    ]
    body = ir.mk_pdf_req(oid, 1, ps, t)
    try:
        im.checkout_pdf(body)
    except Early_Exit:
        pass
    r = im.ses.req
    ro = openapi_core.contrib.requests.RequestsOpenAPIRequest(r)
    oa.validate_request(ro)


def test_buy_png(im, oa):
    im.ses.clear()
    oid = 123
    p = ir.mk_png_pos(1022,
            sender=ir.mk_addr('Erika Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernd Müller', 'Nußhäherstraße 10', '12345', 'München')   )
    t = 100
    body = ir.mk_png_req(oid, p, t)
    try:
        im.checkout_png(body)
    except Early_Exit:
        pass
    r = im.ses.req
    ro = openapi_core.contrib.requests.RequestsOpenAPIRequest(r)
    oa.validate_request(ro)


def test_buy_png_multi(im, oa):
    im.ses.clear()
    oid = 123
    ps = [
        ir.mk_png_pos(1022,
            sender=ir.mk_addr('Erika Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Bernd Müller', 'Nußhäherstraße 10', '12345', 'München')),
        ir.mk_png_pos(1002,
            sender=ir.mk_addr('Peter Mustermann', 'Heidestr. 17', '51147', 'Köln'),
            receiver=ir.mk_addr('Petra Müller', 'Nußhäherstraße 10', '12345', 'München')),
    ]
    t = 100
    body = ir.mk_png_req(oid, ps, t)
    try:
        im.checkout_png(body)
    except Early_Exit:
        pass
    r = im.ses.req
    ro = openapi_core.contrib.requests.RequestsOpenAPIRequest(r)
    oa.validate_request(ro)


def test_preview_png(im, oa):
    im.ses.clear()
    d = ir.mk_png_preview_req(10001)
    try:
        im.preview_png(d)
    except Early_Exit:
        pass
    r = im.ses.req
    ro = openapi_core.contrib.requests.RequestsOpenAPIRequest(r)
    oa.validate_request(ro)


def test_preview_pdf(im, oa):
    im.ses.clear()
    d = ir.mk_pdf_preview_req(5, 1017)
    try:
        im.preview_pdf(d)
    except Early_Exit:
        pass
    r = im.ses.req
    ro = openapi_core.contrib.requests.RequestsOpenAPIRequest(r)
    oa.validate_request(ro)


